package net.sf.freecol.text;

import java.util.EnumMap;
import java.util.List;
import java.util.Random;

public class Unit {
	
	static int expCeiling = 5000;
	static int maxExp = 200;
	static Random r = new Random();
	
	protected String name;
	protected Location loc;
	protected Colony col;
	protected Goods.Types role;
	protected Tile pos;
	protected int exp;
	protected Goods.Types expertise;
	public boolean busy;
	
	protected EnumMap<Goods.Types, Double> thresholds;

	public Unit() {
		this.name = "NONAME";
		this.role = Goods.Types.NONE;
		this.exp = 0;
		this.expertise = Goods.Types.NONE;
		initThresholds();
	}
	
	//unit work in this location
	public Unit(String s, Location l) {
		this.name = s;
		if(l.addUnit(this)) this.loc = l;
		else this.loc = null;
		this.role = l.produces();
		this.exp = 0;
		this.expertise = Goods.Types.NONE;
		initThresholds();
	}
	
	//unit is attached to this colony
	public Unit(String s, Colony c) {
		this.name = s;
		c.addUnit(this);
		this.setLocation(c.outside);
		c.addUnit(this);
		this.role = Goods.Types.NONE;
		this.exp = 0;
		this.expertise = Goods.Types.NONE;
		initThresholds();
	}
	
	public void initThresholds() {
		this.thresholds = new EnumMap<Goods.Types, Double>(Goods.Types.class);
		for(Goods.Types t: Goods.Types.values()) {
			//thresholds start at 1 so the unit is very receptive to new stimuli
			this.thresholds.put(t, 1.0);
		}
	}
	
	public double getThreshold(Goods.Types t) {
		double threshold = 0;
		threshold = this.thresholds.get(t);
		return threshold;
	}
	
	public void setThreshold(Goods.Types t, double d) {
		this.thresholds.put(t, d);
	}
	
	public double sumThresholds() {
		double sum = 0.0;
		for(Goods.Types t: Goods.Types.values()) {
			if(t != Goods.Types.NONE) {
				sum += this.thresholds.get(t);
			}
		}
		return sum;
	}
	
	public String getName() {
		return this.name;
	}
	
	public Location getLocation() {
		return this.loc;
	}
	
	public Colony getColony() {
		return this.col;
	}
	
	public Goods.Types getRole() {
		return this.role;
	}
	
	public void setRole(Goods.Types t) {
		if(this.role != t) {
			//Reset exp
			this.exp = 0;
		}
		this.role = t;
	}
	
	public void setLocation(Location l) {
		//remove unit from old location if needed
		if(this.loc != null) {
			this.loc.clearUnit(this);
		}
		if(l != null) {
			if(l.addUnit(this)) this.loc = l;;
		}
		else {
			this.loc = null;
		}
	}
	
	public void setLocationAndRole(Location l, Goods.Types t) {
		this.setLocation(l);
		//if location doesn't produce any goods of the requested type, change to most profitable good
		if(l.goodsOutput(t)==0) {
			if(l.isBuilding()) {
				this.setRole(l.produces());
			}
			else {
				/**
				 * set to food to be safe but could/should be changed to most profitable good
				 */
				this.setRole(Goods.Types.FOOD);
			}
		}
		else {
			//System.out.println("Unit "+u.name+" has been assigned to "+l.getName()+" for "+u.getRole());
			this.setRole(t);
		}
	}
	
	public boolean assignToBuilding(Colony c, String s) {
		this.col = c;
		col.addUnit(this);
		boolean result = false;
		List<Location> locList = c.getBuildingList();
		for(Location l: locList) {
			if(l.getName().equals(s)) {
				l.addUnit(this);
				this.loc = l;
				this.setRole(l.produces());
				result = true;
				return result;
			}
		}
		this.loc = null;
		return result;
	}
	
	public void setColony(Colony c) {
		this.col = c;
	}
	
	public int produce() {
		
		if (this.loc == null || this.loc == this.col.outside) {
			//System.out.println("This unit has not been assigned to work any location");
			return 0;
		}
		int supply =100;
		if(this.loc.isFactory()) {
			//check fuel supply
			supply = col.storage.get(this.loc.getFuel());
		}
		int production = this.loc.goodsOutput(role);
		//add colony modifier
		production += this.col.getProductionModifier();
		//add expert bonus
		if(this.isExpert(this.getRole())) {
			switch (this.getRole()) {
				case FOOD: production += 3;
					break;
				default : production *= 2;
					break;
			}
		}
		//check if sufficient supplies for factory goods.
		production = Math.min(production, supply);
		//update supplies store
		if(this.loc.isFactory()) {
			this.col.storage.put(this.loc.getFuel(), supply - production);
		}
		return production;
	}
	
	public int produceWithExperience() {
		int prod = this.produce();
		this.updateExperience(prod);
		return prod;
	}
	
	public void updateExperience(int prod) {
		//if(this.getLocation().isTile()) {
			this.exp += prod;
			//cap experience to 200
			this.exp = Math.min(exp, maxExp);
			this.checkPromotion();
			this.adaptThresholdsFromExperience();
		//}
	}
	
	public boolean isExpert(Goods.Types t) {
		return this.expertise.equals(t);
	}
	
	public void printOutput() {
		for(Goods.Types t: Goods.Types.values()) {
			this.setRole(t);
			System.out.println("This unit produces "+produce()+" "+this.role);
		}
	}
	
	public void adaptThresholdsFromExperience() {
	
		for(Goods.Types t : Goods.Types.values()) {
			Double T = this.thresholds.get(t);
			if(this.loc == col.outside) {
				if(!t.equals(expertise)) {
					this.setThreshold(t, Math.max(T-1, 1.0));
					//this.setThreshold(t,  1.0);
				}
			}
			else {
				if(t != Goods.Types.NONE){
					if(t == this.expertise) {
						this.setThreshold(t, 0.1);
					}
					else if (t == this.role){
						//set threshold relative to the amount of exp gained or 0 if not current role
						double threshold = Math.min(10,((maxExp - exp)/20)+1);
						this.setThreshold(t, Math.max(1.0, T-1));
						//this.setThreshold(t, threshold);
					}
					else {
						//this.setThreshold(t, 10.0);
						this.setThreshold(t, Math.min(T+1, 10));
					}
				}
				else this.setThreshold(t, 10.0);
			}
		}
	}
	
	public void checkPromotion() {
		if(this.getLocation().isTile()) {
			int check = r.nextInt(expCeiling);
			if(!this.isExpert(this.getRole())) {
				if(check <= this.exp) {
					//System.out.println("Unit has become expert at "+this.role);
					this.expertise = this.role;
				}
			}
		}
	}
	
	public int expertModifier(Goods.Types t) {
		int modif = 0;
		if(this.isExpert(t)) modif = 3;
		return modif;
	}
	
	public String toString() {
		String s = new String();
		s += "Unit "+this.name+" "+"; expert in "+this.expertise+"; in "+this.loc.getName()
				+" ;producing "+this.role+";threshold "+this.thresholds.get(role)+"\n";
		return s;
	}
	
	public static void main(String[] args) {
		Unit u = new Unit();
		u.role = Goods.Types.FOOD;
		u.adaptThresholdsFromExperience();
		u.exp = 200;
		u.adaptThresholdsFromExperience();
	}
	
}
